import { Doodler, Point, Vector } from "@bearmetal/doodler";
import { PathSegment } from "../math/path.ts";
import { getContextItem, setDefaultContext } from "../lib/context.ts";

export class TrackSystem {
  private segments: Map<string, TrackSegment> = new Map();
  private doodler: Doodler;

  constructor(segments: TrackSegment[]) {
    this.doodler = getContextItem<Doodler>("doodler");
    for (const segment of segments) {
      this.segments.set(segment.id, segment);
    }
  }

  get firstSegment() {
    return this.segments.values().next().value;
  }

  get lastSegment() {
    return this.segments.values().toArray().pop();
  }

  registerSegment(segment: TrackSegment) {
    segment.setTrack(this);
    this.segments.set(segment.id, segment);
  }
  unregisterSegment(segment: TrackSegment) {
    this.segments.delete(segment.id);
    for (const s of this.segments.values()) {
      s.backNeighbours = s.backNeighbours.filter((n) => n !== segment);
      s.frontNeighbours = s.frontNeighbours.filter((n) => n !== segment);
    }
  }

  draw(showControls = false) {
    for (const segment of this.segments.values()) {
      segment.draw(showControls);
    }

    // try {
    //   if (getContextItem<boolean>("showEnds")) {
    //     const ends = this.findEnds();
    //     for (const end of ends) {
    //       this.doodler.fillCircle(end.pos, 2, {
    //         color: "red",
    //         // weight: 3,
    //       });
    //       if (getContextItem<boolean>("debug")) {
    //         this.doodler.line(
    //           end.pos,
    //           end.pos.copy().add(end.tangent.copy().mult(20)),
    //           {
    //             color: "blue",
    //             // weight: 3,
    //           },
    //         );
    //       }
    //     }
    //   }
    // } catch {
    //   setDefaultContext({ showEnds: false });
    // }
  }

  ends: Map<TrackSegment, [End, End]> = new Map();
  endArray: End[] = [];

  findEnds() {
    for (const segment of this.segments.values()) {
      if (this.ends.has(segment)) continue;
      const ends: [End, End] = [
        {
          pos: segment.points[0],
          segment,
          tangent: Vector.sub(segment.points[1], segment.points[0]).normalize(),
          frontOrBack: "back",
        },
        {
          pos: segment.points[3],
          segment,
          tangent: Vector.sub(segment.points[3], segment.points[2]).normalize(),
          frontOrBack: "front",
        },
      ];
      this.ends.set(segment, ends);
      this.endArray.push(...ends);
    }
    return this.endArray;
  }

  serialize() {
    return JSON.stringify(
      this.segments.values().map((s) => s.serialize()).toArray(),
    );
  }

  copy() {
    const track = new TrackSystem([]);
    for (const segment of this.segments.values()) {
      track.segments.set(segment.id, segment.copy());
    }
    return track;
  }

  static deserialize(data: any) {
    if (data.length === 0) return undefined;
    const track = new TrackSystem([]);
    const neighborMap = new Map<string, [string[], string[]]>();
    for (const segment of data) {
      track.segments.set(segment.id, TrackSegment.deserialize(segment));
    }
    for (const segment of track.segments.values()) {
      segment.setTrack(track);
      const neighbors = neighborMap.get(segment.id);
      if (neighbors) {
        segment.backNeighbours = neighbors[1].map((id) =>
          track.segments.get(id)
        ).filter((s) => s) as TrackSegment[];
        segment.frontNeighbours = neighbors[0].map((id) =>
          track.segments.get(id)
        ).filter((s) => s) as TrackSegment[];
      }
    }
    return track;
  }

  translate(v: Vector) {
    for (const segment of this.segments.values()) {
      segment.translate(v);
    }
  }
}

type VectorSet = [Vector, Vector, Vector, Vector];

export class TrackSegment extends PathSegment {
  frontNeighbours: TrackSegment[] = [];
  backNeighbours: TrackSegment[] = [];

  track?: TrackSystem;

  doodler: Doodler;

  id: string;

  constructor(p: VectorSet, id?: string) {
    super(p);
    this.doodler = getContextItem<Doodler>("doodler");
    this.id = id ?? crypto.randomUUID();
  }

  setTrack(t: TrackSystem) {
    this.track = t;
  }

  draw(showControls = false) {
    this.doodler.drawBezier(
      this.points[0],
      this.points[1],
      this.points[2],
      this.points[3],
      {
        strokeColor: "#ffffff50",
      },
    );
    if (showControls) {
      this.doodler.fillCircle(this.points[0], 1, {
        color: "red",
      });
      this.doodler.fillCircle(this.points[1], 1, {
        color: "red",
      });
      this.doodler.fillCircle(this.points[2], 1, {
        color: "red",
      });
      this.doodler.fillCircle(this.points[3], 1, {
        color: "red",
      });
    }
  }

  serialize() {
    return {
      p: this.points.map((p) => p.array()),
      id: this.id,
      bNeighbors: this.backNeighbours.map((n) => n.id),
      fNeighbors: this.frontNeighbours.map((n) => n.id),
    };
  }

  copy() {
    return new TrackSegment(
      this.points.map((p) => p.copy()) as VectorSet,
      this.id,
    );
  }

  cleanCopy() {
    return new TrackSegment(
      this.points.map((p) => p.copy()) as VectorSet,
    );
  }

  propagateTranslation(v: Vector) {
    for (const fNeighbour of this.frontNeighbours) {
      fNeighbour.receivePropagation(v);
    }
    for (const bNeighbour of this.backNeighbours) {
      bNeighbour.receivePropagation(v);
    }
  }

  lastHeading?: number;

  receivePropagation(v: Vector) {
    this.translate(v);
    this.propagateTranslation(v);
  }

  // TODO: this duplicates receivePropagation, but for some reason it doesn't work when called from receivePropagation
  rotate(angle: number | Vector) {
    const [p1, p2, p3, p4] = this.points;
    let newP2;
    if (angle instanceof Vector) {
      const tan = angle;
      angle = tan.heading() - (this.lastHeading ?? 0);
      this.lastHeading = tan.heading();
      newP2 = Vector.add(p1, tan);
    } else {
      const p1ToP2 = Vector.sub(p2, p1);
      p1ToP2.rotate(angle);
      newP2 = Vector.add(p1, p1ToP2);
    }
    const p2ToP3 = Vector.sub(p3, p2);
    p2ToP3.rotate(angle);
    p3.set(Vector.add(newP2, p2ToP3));
    const p2Top4 = Vector.sub(p4, p2);
    p2Top4.rotate(angle);
    p4.set(Vector.add(newP2, p2Top4));
    p2.set(newP2);
  }
  static deserialize(data: any) {
    return new TrackSegment(
      data.p.map((p: [number, number, number]) => new Vector(p[0], p[1], p[2])),
      data.id,
    );
  }

  setPositionByPoint(pos: Vector, point: Vector) {
    if (!this.points.includes(point)) return;
    point = point.copy();
    this.points.forEach((p, i) => {
      const relativePoint = Vector.sub(p, point);
      p.set(pos);
      p.add(relativePoint);
    });
  }

  rotateAboutPoint(angle: number, point: Vector) {
    if (!this.points.includes(point)) return;
    point = point.copy();
    this.points.forEach((p, i) => {
      const relativePoint = Vector.sub(p, point);
      relativePoint.rotate(angle);
      p.set(Vector.add(point, relativePoint));
    });
  }

  // resetRotation() {
  //   const angle = this.tangent(0).heading();
  //   this.rotateAboutPoint(-angle, this.points[0]);
  // }

  translate(v: Point) {
    this.points.forEach((p) => {
      p.add(v.x, v.y);
    });
  }
}
